%%
%%  domain.pl:  Axiomatisation of the "Cooking Agents" domain for MIndiGolog
%%
%%  Author:  Ryan Kelly (rfk)
%
%   Edited by: Xi Liang, Xiaohui Wu and Wade Huang
%
%   Date Edited:    18/05/08
%%
%%  Date Created:  28/07/05
%%
%%    This file contains an axiomatisation of the "Cooking Agents" domain
%%    in the Concurrent, Temporal Situation Calculus with Natural Actions.
%%
%%    The domain consists of several agents and inanimate objects of
%%    different types (indicated by prim_obj/2) which in turn may
%%    be part of super-types (indicated by super_type/2).
%%
%%    Agents may acquire objects, place them inside/on container objects,
%%    and transfer the contents of one container object to another.  There
%%    are also an unlimited supply of timers in the world which may be
%%    set to ring at a specified time in the future.
%%
%%    The agents may also perform several continuous tasks, which have
%%    durations and may span several situations.  Agents may only perform
%%    one task at a time.
%%
%%    Depending on the tasks being performed, the contents of container
%%    objects may evolve from one situation to another.  For example,
%%    if the task mix(bowl1,5) is being performed (mix the contents of
%%    bowl1 for five minutees) then the contents might evolve from
%%    [egg1,flour1,sugar1] to mixed([egg1,flour1,sugar1],5).
%%

%% checkResource(DishList)
%%
%% Check if there is enough resource for the dishes.
%


checkResource(Agt,DishList) :-
    (
        write(DishList),
        addToList(DishList, [], ResNeed),
        getAvailableRes(ResList),
        enoughResource(ResNeed, ResList)
    ->  write('enough resource for cooking'), nl
    ;
        write('not enough resource for cooking'),nl, fail
	 ).


getAvailableRes([tomato,tomato,tomato,tomato,
                 lettuce,lettuce,lettuce,lettuce,lettuce,lettuce,lettuce,lettuce,
                 onion,onion,onion,onion,onion,onion,
                 beef,beef,beef,beef,beef,beef]).	 
					  
					  
resourceList(salad,[lettuce,tomato]).
resourceList(steak,[beef,onion]).

addToList([], ResList, ResList).

addToList([H|T], OldList, ResList)  :-
    resourceList(H, Res),
    append(OldList, Res, NewList),
    addToList(T, NewList, ResList).

enoughResource([H|T], List)  :-
    select(H, List, NewList),
    enoughResource(T, NewList).

enoughResource([], _).


%%  
%%  agent(Agt):  specify agents in the system
%%
%%  This predicate is true when Agt is the name of an agent in the world.
%%
agent(Agt) :-
    prim_agent(Agt,_).



%%
%%  prim_agent(Agt,Type):  specify agents os type Type in the system
%%
%%  This predicate is true when Agt is the name of a primitive agent
%%  in the world, of type Type.
%% 
prim_agent(Agt,chopper) :-
    member(Agt,[chopper1,chopper2,chopper3,chopper4,chopper5,chopper6,chopper7,chopper8,chopper9,chopper10,chopper11,chopper12]).

prim_agent(Agt,assorter) :-
    member(Agt,[assorter1,assorter2,assorter3,assorter4,assorter5,assorter6]).

prim_agent(Agt,roaster) :-
    member(Agt,[roaster1,roaster2,roaster3,roaster4,roaster5,roaster6]).

prim_agent(Agt,dishwasher) :-
    member(Agt,[dishwasher1,dishwasher2,dishwasher3,dishwasher4,dishwasher5,dishwasher6]).

prim_agent(Agt,guest) :-
    member(Agt,[guest1,guest2,guest3,guest4,guest5,guest6]).

prim_agent(Agt, utensilManager)    :-
    member(Agt,[utensilM1,utensilM2,utensilM3]).

prim_agent(Agt, waiter) :-
    member(Agt, [waiter1]).

prim_agent(Agt, resManager) :-
    member(Agt, [resManager1]).



%% 
%%  task(T):  specify the tasks that can be performed
%%
%%  This predicate is true when T is a task that the agents in the system
%%  can perform.  Tasks are typically parameterised in terms of the
%%  objects on which they operate.
%%

%%  assprt(Cont):  assort the contents of container Cont
task(assort(Cont)) :- 
    obj_is_type(Cont,container).

%%  chop(Cont):  chop the contents of container Cont
task(chop(Cont)) :-
    obj_is_type(Cont,container).

%%  roast(Cont):  roast the contents of container Cont
%% task(roast(Cont)) :-
%%    obj_is_type(Cont,container).



%%
%%  task_duration(Agt,Task,Dur):  specify duration of tasks
%%
%%  This predicate is true when Dur is the time taken by agent Agt
%%  to perform the task Task.
%%
task_duration(Agt,assort(_),D) :-
    (Agt = assorter1 ->
        D = 9
    ;
        D = 8
    ).

task_duration(Agt,chop(_),D) :-
    (Agt = chopper1 ->
        D = 5
    ;
        D = 3
    ).


task_duration(Agt,roast(_),D) :-
    (Agt = roaster1 ->
        D = 20
    ;
        D = 15
    ).
	 
%%
%%  prim_obj(Obj,Type):  specify primitive objects in the world
%%
%%  This predicate is true when Obj is the name of a primitive object
%%  in the world, of type Type.
%%
%%  prim_obj(Obj):  shortcut to check object names
%%
%%  This predicate is true if Obj is the name of a primite object,
%%  regardless of its type.
%%

prim_obj(Obj) :-
    prim_obj(Obj,_).


prim_obj(Obj,plate) :-
    member(Obj,[plate1,plate2,plate3,plate4,plate5,plate6,plate7,plate8]).
prim_obj(Obj,board) :-
    member(Obj,[board1,board2,board3,board4,board5,board6,board7,board8,board9,board10,board12,board13,board14,board15,board16,board17,board18,board19]).
prim_obj(Obj,oven) :-
    member(Obj,[oven1, oven2, oven3, oven4, oven5, oven6]).
prim_obj(Obj,sink) :-
    member(Obj,[sink1,sink2,sink3,sink4,sink5,sink6,sink7,sink8,sink9,sink10,sink12,sink13,sink14,sink15,sink16,sink17,sink18,sink19]).
prim_obj(Obj,cupboard) :-
    member(Obj,[cupboard1, cupboard2, cupboard3, cupboard4, cupboard5, cupboard6]).
prim_obj(Obj,mainplate) :-
    member(Obj,[mainplate1,mainplate2,mainplate3,mainplate4,mainplate5,mainplate6,mainplate7,mainplate8]).
prim_obj(Obj,tomato) :-
    member(Obj,[tomato1,tomato2,tomato3,tomato4, tomato5,tomato6,tomato7,tomato8]).
prim_obj(Obj,lettuce) :-
    member(Obj,[lettuce1,lettuce2,lettuce3,lettuce4,lettuce5,lettuce6,lettuce7,lettuce8]).
prim_obj(Obj,carrot) :-
    member(Obj,[carrot1,carrot2,carrot3]).
prim_obj(Obj,onion) :-
    member(Obj,[onion1,onion2,onion3,onion4,onion5,onion6]).
prim_obj(Obj,beef) :-
    member(Obj,[beef1,beef2,beef3,beef4,beef5,beef6]).
	 
   
					  

%%
%%  super_type(SubType,SuperType):  specify type hierarchy
%%
%%  This predicate is true when all objects of type SubType are
%%  also of type SuperType.
%%  
super_type(Type,cooking_appliance) :-
    member(Type,[oven]).
super_type(Type,container) :-
    member(Type,[plate,board,sink,cupboard,mainplate,cooking_appliance]).
super_type(Type,material) :-
    member(Type,[tomato,lettuce,carrot,onion,beef]).

%%
%%  obj_is_type(Obj,Type):  check object types
%%
%%  This predicate is true when the object named Obj is of type
%%  Type according to the hierarchy of super-types.
%%
obj_is_type(Obj,Type) :-
    prim_obj(Obj,Type)
    ;
    super_type(SubType,Type), obj_is_type(Obj,SubType).



%%
%%  prim_action(Act):  specify primitive actions
%%
%%  This predicate is true when Act is the name of a primitive action
%%  in the world.  Actions are typically parameterised in terms of the
%%  objects they act on.  See the details of the MIndiGolog situation
%%  calculus for further information.
%%


    


%%  acquire_object(Agt,Obj):  agent acquires control of an object
prim_action(acquire_object(Agt,Obj)) :-
    agent(Agt), prim_obj(Obj).

%% takeCont(Agt, Obj):  agent take a container.
prim_action(takeCont(Agt, Obj)) :-
    agent(Agt), prim_obj(Obj).


%%  release_object(Agt,Obj):  agent releases an object it has acquired
prim_action(release_object(Agt,Obj)) :-
    agent(Agt), prim_obj(Obj).
	 
prim_action(return_cont(Agt,Obj)) :-
    agent(Agt), prim_obj(Obj).
	 

%%  set_timer(Agt,ID,Dur):  agent sets timer with name ID to ring Dur
%%  minutes in the future
prim_action(set_timer(Agt,_,_)) :-
    agent(Agt).

%%  ring_timer(ID):  timer with name ID rings, a natural action
prim_action(ring_timer(_)).
natural(ring_timer(_)).

%%  place_in(Agt,Conts,Dest):  agent places object Conts in container Dest
prim_action(place_in(Agt,Conts,Dest)) :-
    agent(Agt), prim_obj(Conts), obj_is_type(Dest,container).

prim_action(hold(Agt,Conts)) :-
    agent(Agt), prim_obj(Conts), obj_is_type(Conts,container).


%%  transfer(Agt,Source,Dest):  agent transfers contents of container Source
%%  to container Dest
prim_action(transfer(Agt,Source,Dest)) :-
    agent(Agt), obj_is_type(Source,container),
    obj_is_type(Dest,container).

%%  begin_task(Agt,Task):  agent starts performing task Task
prim_action(begin_task(Agt,Task)) :-
    agent(Agt), task(Task).

%%  end_task(Agt,Task):  agent finishes performing task, a natural action
prim_action(end_task(Agt,Task)) :-
    agent(Agt), task(Task).
natural(end_task(_,_)).


%%
%%  poss(A,T,S):  possibility of performing an action
%%
%%  This predicate is true when it is possible to perform action
%%  A at time T in situation S.
%%

%%  Agents can only acquire an object if no agent already has it,
%%  they arent doing a task, and the object hasnt been used.
poss(acquire_object(Agt,Obj),_,S) :-
    \+ has_object(_,Obj,S), \+ doing_task(Agt,_,_,S), \+ used(Obj,S).

poss(takeCont(Agt, Obj),_,S) :-
    \+ has_cont(_,Obj,S), \+ doing_task(Agt,_,_,S), \+ taken(Obj, S).


%%  Agents may only release objects that they have, when they arent
%%  currently performing a task
poss(release_object(Agt,Obj),_,S) :-
    has_object(Agt,Obj,S), \+ doing_task(Agt,_,_,S).

poss(return_cont(Agt,Obj),_,S) :-
    has_cont(Agt,Obj,S), \+ doing_task(Agt,_,_,S).


%%  Agents may set a timer as long as it hasnt already been set, and
%%  they arent currently performing a task
poss(set_timer(Agt,ID,_),_,S) :-
    \+ timer_set(ID,_,S), \+ doing_task(Agt,_,_,S).

%%  It is only possible for a timer to ring once its remaining time
%%  has precisely elapsed
poss(ring_timer(ID),T,S) :-
    timer_set(ID,Dur,S),
    start(S,SStart), {T = SStart + Dur}.

%%  Agents may place an object in a container provided they have possession
%%  of both, and arent currently doing a task
poss(place_in(Agt,Conts,Dest),_,S) :-
    has_object(Agt,Conts,S), has_object(Agt,Dest,S),
    \+ obj_is_type(Conts,cooking_appliance),
    \+ doing_task(Agt,_,_,S).


poss(hold(Agt,Conts),_,S) :-
    has_cont(Agt,Conts,S),
    \+ obj_is_type(Conts,container),
    \+ doing_task(Agt,_,_,S).


%%  Agents may transfer contents from one container to another as long
%%  as they have possession of both, and arent doing a task
poss(transfer(Agt,Source,Dest),_,S) :-
    has_object(Agt,Source,S), has_object(Agt,Dest,S),
    \+ doing_task(Agt,_,_,S).

%%  Agents may begin the assort() task as long as they arent doing another
%%  task, and have possession of the container to be mixed in
poss(begin_task(Agt,assort(Obj)),_,S) :-
    has_object(Agt,Obj,S), \+ doing_task(Agt,_,_,S).

%%  Agents may begin the chop() task as long as they arent doing another
%%  task, and have possession of the container whose contents to chop
poss(begin_task(Agt,chop(Obj)),_,S) :-
    has_object(Agt,Obj,S), \+ doing_task(Agt,_,_,S).

%%  Agents may begin the roast() task as long as they arent doing another
%%  task, and have possession of the container whose contents to chop
poss(begin_task(Agt,roast(Obj)),_,S) :-
    has_object(Agt,Obj,S), \+ doing_task(Agt,_,_,S).


%%  Agents may end a task only when precisely the remaining amount of
%%  time has elapsed
poss(end_task(Agt,Task),T,S) :-
    doing_task(Agt,Task,Remain,S),
    start(S,SStart), {T = SStart + Remain}.


%%
%%  conflicts(C,T,S):  concurrent actions conflict
%%
%%  This predicate must be true when concurrent actions in C conflict
%%  and cannot be performed at time T in situation S.
%%

%%  Agents cannot do more than one action at a time
conflicts(C,_,_) :-
    member(A1,C), actor(A1,Agt),
    member(A2,C), actor(A2,Agt),
    A2 \= A1.

%%  Two agents cannot acquire the same object
conflicts(C,_,_) :-
    member(acquire_object(A1,Res),C),
    member(acquire_object(A2,Res),C),
    A1 \= A2.
    
%%
%%  Fluents in the Domain
%%
%%  The fluents are specified in terms of their successor state axioms,
%%  of the form "a fluent is true if it became true, or was previously
%%  true did not become false".
%%
%%    fluent_holds(Args,do(A,T,S)) :-
%%        fluent_becomes_true(Args,do(A,T,S))
%%        ;
%%        (
%%          fluent_holds(Args,S),
%%          \+ fluent_becomes_false(Args,do(A,T,S))
%%        )
%%

%%
%%  has_object(Agt,Obj,S):  agent has an object
%%
%%  This fluent is true when agent Agt has possession of the object Obj
%%  in situation S.  It can become true by acquiring the object, and
%%  false by releasing the object or if it has become used.
%%
has_object(Agt,Obj,do(C,T,S)) :-
    member(acquire_object(Agt,Obj),C)
    ;
    has_object(Agt,Obj,S),
    \+ (
       member(release_object(Agt,Obj),C)
       ;
       used(Obj,do(C,T,S))
    ).

has_cont(Agt, Obj,do(C,T,S)):-
    member(takeCont(Agt, Obj),C)
    ;
    has_cont(Agt, Obj,S),
    \+ (
        member(return_cont(Agt, Obj),C)
        ;
        taken(Obj,do(C,T,S))
    ).


%%
%%  used(Obj,S):  object is used in situation S
%%
%%  This fluent is true when an object has been used - for example,
%%  an ingredient has been placed in a container.  Once an object has
%%  been used, it cannot be used again.
%%
used(Obj,do(C,_,S)) :-
    prim_obj(Obj), obj_is_type(Obj,ingredient),
    (
      used(Obj,S)
      ;
      member(place_in(_,Obj,_),C)
    ).


%%
%   taken(Obj, do(C,_S))    object is taken in situation S
%
%   This fluent is true when an oject has been taken by an utensil manager
%
taken(Obj,do(C,_,S)) :-
    prim_obj(Obj), obj_is_type(Obj,container),
    (
      taken(Obj,S)
      ;
      member(hold(_,Obj,_),C)
    ).


%%
%%  timer_set(ID,Dur,S):  timer is set in situation S
%%
%%  This fluent indicates that the timer named ID is set to ring in Dur
%%  minutes in situation S.  It becomes true as a result of a set_timer()
%%  action, and is updated from situation to situation to reflect
%%  the amount of time remaining.
%%
timer_set(ID,Dur,do(C,T,S)) :-
    member(set_timer(_,ID,Dur),C)
    ;
    timer_set(ID,OldDur,S), start(S,SStart), {Dur = OldDur-(T-SStart)},
    \+ member(ring_timer(ID),C).

%%
%%  contents(Obj,Conts,S):  object contents in a situation
%%
%%  This fluent indicates that object Obj contains the contents Conts
%%  in situation S.  It can become true, become false, and change value
%%  in a variety of ways, each of which is documented with its
%%  implementation.
%%
contents(Obj,Conts,do(C,T,S)) :-
    start(S,SStart),
    ((
      %% --- All the ways it can become true
      %% It was previously empty, and contents were placed or transfered in
      (member(place_in(_,Conts,Obj),C)
         ; member(transfer(_,Obj2,Obj),C), contents(Obj2,Conts,S)),
      \+ contents(Obj,_,S)
      ;
      %% It previously had contents, and more contents were placed or
      %% transfered in.  Contents is then a list.
      (member(place_in(_,NewConts,Obj),C)
         ; member(transfer(_,Obj2,Obj),C), contents(Obj2,NewConts,S)),
      contents(Obj,OldConts,S),
      ( OldConts = [_|_] -> OldContsL = OldConts ; OldContsL = [OldConts]),
      ( NewConts = [_|_] -> NewContsL = NewConts ; NewContsL = [NewConts]),
      union(OldContsL,NewContsL,Conts)
      ;

      doing_task(_,assort(Obj),_,do(C,T,S)), contents(Obj,OldConts,S),
      (  OldConts = assorted(MixConts,OldP) ->
             {NewP = OldP+(T-SStart)}, Conts = assorted(MixConts,NewP)
         ;
             Conts = assorted(OldConts,0)
      )
      ;

      member(end_task(_,assort(Obj)),C), contents(Obj,assorted(MixConts,OldP),S),
      {NewP = OldP+(T-SStart)}, Conts = assorted(MixConts,NewP)
      ;
      %% An agent just completed chopping the contents.  They are
      %% encased in a chopped() indicator.
      member(end_task(_,chop(Obj)),C), contents(Obj,OldConts,S),
      Conts = chopped(OldConts)
      ;
      %% If the container is in an oven, its contents are baked.
      %% If they are not encapsulated in a baked() indicator then do
      %% so, otherwise update the baking time.
      \+ obj_is_type(Obj,cooking_appliance), obj_is_type(Oven,oven),
      contents(Oven,Obj,do(C,T,S)), contents(Obj,OldConts,S),
      (  OldConts = baked(BakedConts,OldP) ->
             {NewP = OldP+(T-SStart)}, Conts = baked(BakedConts,NewP)
         ;
             Conts = baked(OldConts,0)
      )
      ;
      %% If the container was just taken out of the oven, updated
      %% the content to reflect the total baking time.
      \+ obj_is_type(Obj,cooking_appliance), obj_is_type(Oven,oven),
      contents(Oven,Obj,S), member(transfer(_,Oven,_),C),
      contents(Obj,baked(BakedConts,OldP),S),
      {NewP = OldP+(T-SStart)}, Conts = baked(BakedConts,NewP)
    )
    ;
    %% Or it was true, and didnt become false...
    contents(Obj,Conts,S), \+ (
        %% --- All the ways it can become false
        %% The contents were transfered out
        member(transfer(_,Obj,_),C)
        ;
        %% New contents were transfered in
        member(transfer(_,Obj2,Obj),C), contents(Obj2,_,S)
        ;
        %% New contents were placed in
        member(place_in(_,_,Obj),C)
        ;
        doing_task(_,assort(Obj),_,do(C,T,S))
        ;
        %% The contents just finished being mixed, hence will change
        member(end_task(_,assort(Obj)),C)
        ;
        %% The contents just finished being chopped, hence will change
        member(end_task(_,chop(Obj)),C)
        ;
        %% The object is in an oven, hence will change
        \+ obj_is_type(Obj,cooking_appliance), obj_is_type(Oven,oven),
        contents(Oven,Obj,do(C,T,S))
        ;
        %% The object was just taken out of an oven, hence will change
        \+ obj_is_type(Obj,cooking_appliance), obj_is_type(Oven,oven),
        contents(Oven,Obj,S), member(transfer(_,Oven,_),C)
    )).


%%
%%  doing_task(Agt,Task,Remain,S):  agent is performing a task
%%
%%  This fluent is true when agent Agt is performing task Task in situation
%%  S, and has Remain minutes left before it is completed.  It becomes
%%  true when an agent begins a task, becomes false when an agent
%%  completes a task, and its value is updated from situation to situation
%%  to reflect the remianing time till completion.
%%
doing_task(Agt,Task,Remain,do(C,T,S)) :-
    member(begin_task(Agt,Task),C), task_duration(Agt,Task,Remain)
    ;
    doing_task(Agt,Task,OldRem,S), start(S,SStart),
    {OldRem = Remain-(T-SStart)}, \+ member(end_task(Agt,Task),C).
    

%%
%%  history_length(N,S):  length of the action histoy in a situation
%%
%%  This simple fluent encodes in N the number of actions that have
%%  taken place in the history of situation S.  It is used to make this
%%  information easily available to agents.
%%
history_length(N,do(_,_,S)) :-
    history_length(N1,S),
    N is N1 + 1.
history_length(0,s0).

%%
%%  Intial Conditions for the domain
%%
%%  The initial conditions are specified by additional clauses for
%%  each fluent, with the situation term set to the atom s0.  For
%%  the most part no fluents hold in the initial situation, so 
%%  there arent many clauses here.
%%

start(s0,0).


